@@ -1243,7 +1243,7 @@ void TypeInst::mkPar(EnvI& env) {
   }
 }
 
-void TypeInst::setStructDomain(const EnvI& env, const Type& struct_type, bool setTypeAny,
+void TypeInst::setStructDomain(EnvI& env, const Type& struct_type, bool setTypeAny,
                                bool setTIRanges) {
   GCLock lock;
   StructType* st = env.getStructType(struct_type);
@@ -1275,7 +1275,20 @@ void TypeInst::setStructDomain(const EnvI& env, const Type& struct_type, bool se
   auto* al = ArrayLit::constructTuple(Expression::loc(this).introduce(), field_ti);
   al->type(struct_type.elemType(env));
   domain(al);
-  type(struct_type);
+  if (struct_type.dim() == 0 && !ranges().empty()) {
+    std::vector<unsigned int> enumIds(ranges().size() + 1);
+    for (unsigned int k = 0; k < ranges().size(); k++) {
+      enumIds[k] = ranges()[k]->type().typeId();
+    }
+    enumIds[ranges().size()] = struct_type.typeId();
+    Type t = struct_type;
+    t.typeId(0);
+    t.dim(static_cast<int>(ranges().size()));
+    t.typeId(env.registerArrayEnum(enumIds));
+    type(t);
+  } else {
+    type(struct_type);
+  }
 }
 
 bool TypeInst::resolveAlias(EnvI& env) {
@@ -1321,7 +1334,8 @@ bool TypeInst::resolveAlias(EnvI& env) {
     }
     ntype.st(Type::ST_SET);
   }
-  assert(type().dim() == ranges().size() && ntype.dim() == alias->ranges().size());
+  assert(type().dim() == -1 ||
+         type().dim() == ranges().size() && ntype.dim() == alias->ranges().size());
   if (type().dim() != 0) {
     if (ntype.dim() != 0) {
       std::stringstream ss;
@@ -1330,7 +1344,7 @@ bool TypeInst::resolveAlias(EnvI& env) {
          << "' and is already an array type";
       throw TypeError(env, Expression::loc(this), ss.str());
     }
-    const int dim = type().dim();
+    const int dim = type().dim() == -1 ? 1 : type().dim();
     const unsigned int curTypeId = type().typeId();
     const unsigned int newTypeId = ntype.typeId();
     if (curTypeId != 0 || newTypeId != 0) {
@@ -1347,10 +1361,10 @@ bool TypeInst::resolveAlias(EnvI& env) {
         arrayEnumIds[dim] = newTypeId;
       }
       ntype.typeId(0);
-      ntype.dim(dim);
+      ntype.dim(type().dim());
       ntype.typeId(env.registerArrayEnum(arrayEnumIds));
     } else {
-      ntype.dim(dim);
+      ntype.dim(type().dim());
     }
   } else if (ntype.dim() != 0) {
     std::vector<TypeInst*> ranges(alias->ranges().size());
@@ -1404,6 +1418,11 @@ bool TypeInst::concatDomain(EnvI& env) {
   }
   // Update TI
   domain(dom);
+
+  unsigned int tId = ty.typeId();
+  ty.typeId(0);
+  ty.dim(type().dim());
+  ty.typeId(tId);
   type(ty);
   return true;
 }
@@ -1437,7 +1456,10 @@ const Location& get_loc(Expression* e, Expression* call, FunctionI* /*fi*/) {
   return Expression::loc(e).isNonAlloc() ? Expression::loc(call) : Expression::loc(e);
 }
 const Location& get_loc(const Type& /*t*/, Expression* call, FunctionI* fi) {
-  return Expression::loc(call).isNonAlloc() ? fi->loc() : Expression::loc(call);
+  if (call == nullptr || Expression::loc(call).isNonAlloc()) {
+    return fi == nullptr ? Location::nonalloc : fi->loc();
+  }
+  return Expression::loc(call);
 }
 
 bool isa_tiid(Expression* e) {
@@ -1550,11 +1572,7 @@ Type return_type(EnvI& env, FunctionI* fi, const std::vector<T>& ta, Expression*
       }
       Type tiit = Type::top(orig_tiit.dim());
       if (orig_tiit.typeId() != 0) {
-        std::vector<unsigned int> enumIds(tiit.dim() + 1);
-        const std::vector<unsigned int>& orig_enumIds = env.getArrayEnum(orig_tiit.typeId());
-        for (unsigned int i = 0; i < enumIds.size() - 1; i++) {
-          enumIds[i] = orig_enumIds[i];
-        }
+        std::vector<unsigned int> enumIds = env.getArrayEnum(orig_tiit.typeId());
         enumIds[enumIds.size() - 1] = 0;
         tiit.typeId(env.registerArrayEnum(enumIds));
       }
@@ -1801,11 +1819,15 @@ Type FunctionI::argtype(EnvI& env, const std::vector<Expression*>& ta, unsigned
   // least common supertype that fits function parameter n.
   TypeInst* tii = param(n)->ti();
   Type curTiiT = tii->type();
+  Type dimTy = curTiiT;
   if (curTiiT.dim() == -1) {
     if (Expression::type(ta[n]).dim() == 0) {
-      curTiiT.dim(1);
+      dimTy = Type::partop(1);
     } else {
-      curTiiT.dim(Expression::type(ta[n]).dim());
+      dimTy = Expression::type(ta[n]);
+      if (dimTy.dim() == -1) {
+        dimTy = Type::partop(1);
+      }
     }
   }
   if ((tii->domain() != nullptr) && Expression::isa<TIId>(tii->domain())) {
@@ -1818,11 +1840,11 @@ Type FunctionI::argtype(EnvI& env, const std::vector<Expression*>& ta, unsigned
     if (!ty.structBT()) {
       ty.st(curTiiT.st());
     }
-    if (curTiiT.dim() != ty.dim()) {
-      if (curTiiT.dim() == 0) {
+    if (dimTy.dim() != ty.dim()) {
+      if (dimTy.dim() == 0) {
         ty = ty.elemType(env);
       } else {
-        ty = Type::arrType(env, curTiiT.dim() > 0 ? curTiiT : Type::partop(1), ty);
+        ty = Type::arrType(env, dimTy, ty);
       }
     }
     ASTString tv = Expression::cast<TIId>(tii->domain())->v();
@@ -1835,11 +1857,11 @@ Type FunctionI::argtype(EnvI& env, const std::vector<Expression*>& ta, unsigned
           toCheck.ot(curTiiT.ot());
           toCheck.st(curTiiT.st());
         }
-        if (curTiiT.dim() != toCheck.dim()) {
-          if (curTiiT.dim() == 0) {
+        if (dimTy.dim() != toCheck.dim()) {
+          if (dimTy.dim() == 0) {
             toCheck = toCheck.elemType(env);
           } else {
-            toCheck = Type::arrType(env, curTiiT.dim() > 0 ? curTiiT : Type::partop(1), toCheck);
+            toCheck = Type::arrType(env, dimTy, toCheck);
           }
         }
         if (toCheck != ty) {
@@ -2175,6 +2197,7 @@ Constants::Constants() {
   ids.join = addString("join");
   ids.show = addString("show");
   ids.format = addString("format");
+  ids.format_justify_string = addString("format_justify_string");
   ids.showDzn = addString("showDzn");
   ids.showJSON = addString("showJSON");
   ids.output = addString("output");
@@ -2368,6 +2391,8 @@ Constants::Constants() {
   ann.set_search = addString("set_search");
   ann.warm_start = addString("warm_start");
   ann.warm_start_array = addString("warm_start_array");
+  ann.computed_domain = addId("computed_domain");
+  ann.computed_domain->type(Type::ann());
 
   cli.cmdlineData_short_str = addString("-D");
   cli.cmdlineData_str = addString("--cmdline-data");
