@@ -830,7 +830,10 @@ void flatten_linexp_binop(EnvI& env, const Ctx& ctx, VarDecl* r, VarDecl* b, EE&
             ret.r = bind(env, ctx, r, env.constants.literalFalse);
           }
         } else {
-          set_computed_domain(env, vd, LinearTraits<Lit>::newDomain(bounds.l, bounds.u), true);
+          // Can only set as computed if there was no other RHS
+          bool is_computed = vd->e() == nullptr;
+          set_computed_domain(env, vd, LinearTraits<Lit>::newDomain(bounds.l, bounds.u),
+                              is_computed);
         }
       }
     }
@@ -1291,7 +1294,7 @@ EE flatten_bool_op(EnvI& env, Ctx& ctx, const Ctx& ctx0, const Ctx& ctx1, Expres
   } else if (isBuiltin && bot == BOT_EQ && ctx.b == C_ROOT && r == env.constants.varTrue &&
              Expression::isa<Id>(e0.r()) && Expression::type(e1.r()).isPar() &&
              !env.hasReverseMapper(Expression::cast<Id>(e0.r()))) {
-    // Par assignment to Id with no right hand side
+    // Par assignment to Id
     auto* vd = Expression::cast<Id>(e0.r())->decl();
     (void)bind(env, ctx, vd, e1.r());
     ees.resize(2);
@@ -1526,6 +1529,19 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
   Expression* boe0 = bo->lhs();
   Expression* boe1 = bo->rhs();
   bool isBuiltin = bo->decl() == nullptr || bo->decl()->e() == nullptr;
+
+  if (ctx.neg) {
+    if (bot == BOT_EQUIV) {
+      ctx.neg = false;
+      ctx.b = -ctx.b;
+      bot = BOT_XOR;
+    } else if (bot == BOT_XOR) {
+      ctx.neg = false;
+      ctx.b = -ctx.b;
+      bot = BOT_EQUIV;
+    }
+  }
+
   switch (bot) {
     case BOT_AND:
       if (isBuiltin) {
@@ -1747,27 +1763,19 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
       }
     } break;
     case BOT_EQUIV:
-      if (ctx.neg) {
-        ctx.neg = false;
-        ctx.b = -ctx.b;
-        bot = BOT_XOR;
-        ctx0.b = ctx1.b = C_MIX;
-        return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
-      }
+      assert(!ctx.neg);
       if (!Expression::type(boe1).isOpt() && istrue(env, boe0)) {
         return flat_exp(env, ctx, boe1, r, b);
       }
       if (!Expression::type(boe0).isOpt() && istrue(env, boe1)) {
         return flat_exp(env, ctx, boe0, r, b);
       }
-      if (!Expression::type(boe0).isOpt() && !Expression::type(boe1).isOpt() && (r != nullptr) &&
+      if (!Expression::type(boe0).isOpt() && !Expression::type(boe1).isOpt() &&
           r == env.constants.varTrue) {
         if (Expression::type(boe1).isPar() || Expression::isa<Id>(boe1)) {
           std::swap(boe0, boe1);
         }
-        if (istrue(env, boe0)) {
-          return flat_exp(env, ctx1, boe1, r, b);
-        }
+        assert(!istrue(env, boe0));
         if (isfalse(env, boe0)) {
           ctx1.neg = true;
           ctx1.b = -ctx1.b;
@@ -1794,10 +1802,42 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
       ctx0.b = ctx1.b = C_MIX;
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
     case BOT_XOR:
-      if (ctx.neg) {
-        ctx.neg = false;
-        ctx.b = -ctx.b;
-        bot = BOT_EQUIV;
+      assert(!ctx.neg);
+      if (!Expression::type(boe1).isOpt() && isfalse(env, boe0)) {
+        return flat_exp(env, ctx, boe1, r, b);
+      }
+      if (!Expression::type(boe0).isOpt() && isfalse(env, boe1)) {
+        return flat_exp(env, ctx, boe0, r, b);
+      }
+      if (!Expression::type(boe0).isOpt() && !Expression::type(boe1).isOpt() &&
+          r == env.constants.varTrue) {
+        if (Expression::type(boe1).isPar() || Expression::isa<Id>(boe1)) {
+          std::swap(boe0, boe1);
+        }
+        assert(!isfalse(env, boe0));
+        if (istrue(env, boe0)) {
+          ctx1.neg = true;
+          ctx1.b = -ctx1.b;
+          return flat_exp(env, ctx1, boe1, r, b);
+        }
+        ctx0.b = C_MIX;
+        EE e0 = flat_exp(env, ctx0, boe0, nullptr, ctx0.partialityVar(env));
+        if (isfalse(env, e0.r())) {
+          return flat_exp(env, ctx1, boe1, r, b);
+        }
+        if (istrue(env, e0.r())) {
+          ctx1.neg = true;
+          ctx1.b = -ctx1.b;
+          return flat_exp(env, ctx1, boe1, r, b);
+        }
+        Id* id = Expression::cast<Id>(e0.r());
+        ctx1.neg = true;
+        ctx1.b = C_MIX;
+        (void)flat_exp(env, ctx1, boe1, id->decl(), env.constants.varTrue);
+        env.addCtxAnn(id->decl(), ctx1.b);
+        ret.b = bind(env, Ctx(), b, env.constants.literalTrue);
+        ret.r = bind(env, Ctx(), r, env.constants.literalTrue);
+        break;
       }
       ctx0.b = ctx1.b = C_MIX;
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
@@ -1805,84 +1845,52 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
       if (ctx.neg) {
         doubleNeg = true;
         bot = BOT_GQ;
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = +ctx0.b;
-          ctx1.b = -ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = +ctx0.b;
-          ctx1.i = -ctx1.b;
-        }
-      } else {
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = -ctx0.b;
-          ctx1.b = +ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = -ctx0.b;
-          ctx1.i = +ctx1.b;
-        }
+      }
+      if (Expression::type(boe0).bt() == Type::BT_BOOL) {
+        ctx0.b = -ctx0.b;
+        ctx1.b = +ctx1.b;
+      } else if (Expression::type(boe0).bt() == Type::BT_INT) {
+        ctx0.i = -ctx0.b;
+        ctx1.i = +ctx1.b;
       }
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
     case BOT_LQ:
       if (ctx.neg) {
         doubleNeg = true;
         bot = BOT_GR;
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = +ctx0.b;
-          ctx1.b = -ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = +ctx0.b;
-          ctx1.i = -ctx1.b;
-        }
-      } else {
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = -ctx0.b;
-          ctx1.b = +ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = -ctx0.b;
-          ctx1.i = +ctx1.b;
-        }
+      }
+      if (Expression::type(boe0).bt() == Type::BT_BOOL) {
+        ctx0.b = -ctx0.b;
+        ctx1.b = +ctx1.b;
+      } else if (Expression::type(boe0).bt() == Type::BT_INT) {
+        ctx0.i = -ctx0.b;
+        ctx1.i = +ctx1.b;
       }
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
     case BOT_GR:
       if (ctx.neg) {
         doubleNeg = true;
         bot = BOT_LQ;
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = -ctx0.b;
-          ctx1.b = +ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = -ctx0.b;
-          ctx1.i = +ctx1.b;
-        }
-      } else {
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = +ctx0.b;
-          ctx1.b = -ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = +ctx0.b;
-          ctx1.i = -ctx1.b;
-        }
+      }
+      if (Expression::type(boe0).bt() == Type::BT_BOOL) {
+        ctx0.b = +ctx0.b;
+        ctx1.b = -ctx1.b;
+      } else if (Expression::type(boe0).bt() == Type::BT_INT) {
+        ctx0.i = +ctx0.b;
+        ctx1.i = -ctx1.b;
       }
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
     case BOT_GQ:
       if (ctx.neg) {
         doubleNeg = true;
         bot = BOT_LE;
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = -ctx0.b;
-          ctx1.b = +ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = -ctx0.b;
-          ctx1.i = +ctx1.b;
-        }
-      } else {
-        if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-          ctx0.b = +ctx0.b;
-          ctx1.b = -ctx1.b;
-        } else if (Expression::type(boe0).bt() == Type::BT_INT) {
-          ctx0.i = +ctx0.b;
-          ctx1.i = -ctx1.b;
-        }
+      }
+      if (Expression::type(boe0).bt() == Type::BT_BOOL) {
+        ctx0.b = +ctx0.b;
+        ctx1.b = -ctx1.b;
+      } else if (Expression::type(boe0).bt() == Type::BT_INT) {
+        ctx0.i = +ctx0.b;
+        ctx1.i = -ctx1.b;
       }
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
     case BOT_EQ:
@@ -1890,9 +1898,7 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
         doubleNeg = true;
         bot = BOT_NQ;
       }
-      if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-        ctx0.b = ctx1.b = C_MIX;
-      } else if (Expression::type(boe0).bt() == Type::BT_INT) {
+      if (Expression::type(boe0).bt() == Type::BT_INT) {
         ctx0.i = ctx1.i = C_MIX;
       }
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
@@ -1901,9 +1907,7 @@ EE flatten_binop(EnvI& env, const Ctx& input_ctx, Expression* e, VarDecl* r, Var
         doubleNeg = true;
         bot = BOT_EQ;
       }
-      if (Expression::type(boe0).bt() == Type::BT_BOOL) {
-        ctx0.b = ctx1.b = C_MIX;
-      } else if (Expression::type(boe0).bt() == Type::BT_INT) {
+      if (Expression::type(boe0).bt() == Type::BT_INT) {
         ctx0.i = ctx1.i = C_MIX;
       }
       return flatten_bool_op(env, ctx, ctx0, ctx1, e, r, b, isBuiltin, bo, bot, doubleNeg);
