@@ -242,7 +242,7 @@ void unify(EnvI& env, std::vector<VarDecl*>& deletedVarDecls, Id* id0, Id* id1)
             IntSetRanges isv1r(isv1);
             Ranges::Inter<IntVal, IntSetRanges, IntSetRanges> inter(isv0r, isv1r);
             IntSetVal* nd = IntSetVal::ai(inter);
-            if (nd->empty()) {
+            if (nd->empty() && !ti0->type().isSet()) {
               env.fail();
             } else if (!nd->equal(isv1)) {
               ti1->domain(new SetLit(Location(), nd));
@@ -349,6 +349,72 @@ void push_dependent_constraints(EnvI& env, Id* id, std::deque<Item*>& q) {
   }
 }
 
+void remove_deleted_items(EnvI& envi, std::vector<VarDecl*>& deletedVarDecls) {
+  // TODO: The delayed deletion could be done eagerly by the creation of
+  // env.optRemoveItem() which contains the logic in this while loop.
+  auto& m = *envi.flat();
+  while (!deletedVarDecls.empty()) {
+    envi.checkCancel();
+    VarDecl* cur = deletedVarDecls.back();
+    deletedVarDecls.pop_back();
+    if (envi.varOccurrences.occurrences(cur) == 0) {
+      auto cur_idx = envi.varOccurrences.idx.find(cur->id());
+      if (cur_idx.first && !m[*cur_idx.second]->removed()) {
+        if (is_output(cur)) {
+          // We have to change the output model if we remove this variable
+          Expression* val = nullptr;
+          if (cur->type().isbool() && (cur->ti()->domain() != nullptr)) {
+            val = cur->ti()->domain();
+          } else if (cur->type().isint()) {
+            if ((cur->e() != nullptr) && Expression::isa<IntLit>(cur->e())) {
+              val = cur->e();
+            } else if ((cur->ti()->domain() != nullptr) &&
+                       Expression::isa<SetLit>(cur->ti()->domain()) &&
+                       Expression::cast<SetLit>(cur->ti()->domain())->isv()->size() == 1 &&
+                       Expression::cast<SetLit>(cur->ti()->domain())->isv()->min() ==
+                           Expression::cast<SetLit>(cur->ti()->domain())->isv()->max()) {
+              val = IntLit::a(Expression::cast<SetLit>(cur->ti()->domain())->isv()->min());
+            }
+          } else if (cur->type().isfloat()) {
+            if ((cur->e() != nullptr) && Expression::isa<FloatLit>(cur->e())) {
+              val = cur->e();
+            } else if ((cur->ti()->domain() != nullptr) &&
+                       Expression::isa<SetLit>(cur->ti()->domain()) &&
+                       Expression::cast<SetLit>(cur->ti()->domain())->fsv()->size() == 1 &&
+                       Expression::cast<SetLit>(cur->ti()->domain())->fsv()->min() ==
+                           Expression::cast<SetLit>(cur->ti()->domain())->fsv()->max()) {
+              val = FloatLit::a(Expression::cast<SetLit>(cur->ti()->domain())->fsv()->min());
+            }
+          } else if (cur->type().isIntSet()) {
+            if (cur->e() != nullptr && Expression::isa<SetLit>(cur->e())) {
+              val = cur->e();
+            }
+          } else if (cur->type().dim() > 0 && cur->type().isPar()) {
+            if (cur->e() != nullptr && Expression::isa<ArrayLit>(cur->e())) {
+              val = cur->e();
+            }
+          }
+          if (val != nullptr) {
+            // Find corresponding variable in output model and fix it
+            VarDecl* vd_out =
+                (*envi.output)[envi.outputFlatVarOccurrences.find(cur)]->cast<VarDeclI>()->e();
+            vd_out->e(val);
+            CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls,
+                            m[*cur_idx.second]->cast<VarDeclI>());
+            top_down(cd, cur->e());
+            (*envi.flat())[*cur_idx.second]->remove();
+          }
+        } else {
+          CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls,
+                          m[*cur_idx.second]->cast<VarDeclI>());
+          top_down(cd, cur->e());
+          (*envi.flat())[*cur_idx.second]->remove();
+        }
+      }
+    }
+  }
+}
+
 void optimize(Env& env, bool chain_compression) {
   env.envi().checkCancel();
 
@@ -542,7 +608,8 @@ void optimize(Env& env, bool chain_compression) {
         }
         if (vdi->e()->type().isbool() && vdi->e()->type().dim() == 0 &&
             (vdi->e()->ti()->domain() == envi.constants.literalTrue ||
-             vdi->e()->ti()->domain() == envi.constants.literalFalse)) {
+             vdi->e()->ti()->domain() == envi.constants.literalFalse ||
+             (vdi->e()->e() != nullptr && Expression::isa<BoolLit>(vdi->e()->e())))) {
           // push RHS onto constraint queue since this bool var is fixed
           push_vardecl(envi, vdi, i, vardeclQueue);
           push_dependent_constraints(envi, vdi->e()->id(), constraintQueue);
@@ -1062,75 +1129,16 @@ void optimize(Env& env, bool chain_compression) {
     }
 
     // Phase 6: remove deleted variables if possible
-    // TODO: The delayed deletion could be done eagerly by the creation of
-    // env.optRemoveItem() which contains the logic in this while loop.
-    while (!deletedVarDecls.empty()) {
-      env.envi().checkCancel();
-      VarDecl* cur = deletedVarDecls.back();
-      deletedVarDecls.pop_back();
-      if (envi.varOccurrences.occurrences(cur) == 0) {
-        auto cur_idx = envi.varOccurrences.idx.find(cur->id());
-        if (cur_idx.first && !m[*cur_idx.second]->removed()) {
-          if (is_output(cur)) {
-            // We have to change the output model if we remove this variable
-            Expression* val = nullptr;
-            if (cur->type().isbool() && (cur->ti()->domain() != nullptr)) {
-              val = cur->ti()->domain();
-            } else if (cur->type().isint()) {
-              if ((cur->e() != nullptr) && Expression::isa<IntLit>(cur->e())) {
-                val = cur->e();
-              } else if ((cur->ti()->domain() != nullptr) &&
-                         Expression::isa<SetLit>(cur->ti()->domain()) &&
-                         Expression::cast<SetLit>(cur->ti()->domain())->isv()->size() == 1 &&
-                         Expression::cast<SetLit>(cur->ti()->domain())->isv()->min() ==
-                             Expression::cast<SetLit>(cur->ti()->domain())->isv()->max()) {
-                val = IntLit::a(Expression::cast<SetLit>(cur->ti()->domain())->isv()->min());
-              }
-            } else if (cur->type().isfloat()) {
-              if ((cur->e() != nullptr) && Expression::isa<FloatLit>(cur->e())) {
-                val = cur->e();
-              } else if ((cur->ti()->domain() != nullptr) &&
-                         Expression::isa<SetLit>(cur->ti()->domain()) &&
-                         Expression::cast<SetLit>(cur->ti()->domain())->fsv()->size() == 1 &&
-                         Expression::cast<SetLit>(cur->ti()->domain())->fsv()->min() ==
-                             Expression::cast<SetLit>(cur->ti()->domain())->fsv()->max()) {
-                val = FloatLit::a(Expression::cast<SetLit>(cur->ti()->domain())->fsv()->min());
-              }
-            } else if (cur->type().isIntSet()) {
-              if (cur->e() != nullptr && Expression::isa<SetLit>(cur->e())) {
-                val = cur->e();
-              }
-            } else if (cur->type().dim() > 0 && cur->type().isPar()) {
-              if (cur->e() != nullptr && Expression::isa<ArrayLit>(cur->e())) {
-                val = cur->e();
-              }
-            }
-            if (val != nullptr) {
-              // Find corresponding variable in output model and fix it
-              VarDecl* vd_out =
-                  (*envi.output)[envi.outputFlatVarOccurrences.find(cur)]->cast<VarDeclI>()->e();
-              vd_out->e(val);
-              CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls,
-                              m[*cur_idx.second]->cast<VarDeclI>());
-              top_down(cd, cur->e());
-              (*envi.flat())[*cur_idx.second]->remove();
-            }
-          } else {
-            CollectDecls cd(envi, envi.varOccurrences, deletedVarDecls,
-                            m[*cur_idx.second]->cast<VarDeclI>());
-            top_down(cd, cur->e());
-            (*envi.flat())[*cur_idx.second]->remove();
-          }
-        }
-      }
-    }
-  } catch (ModelInconsistent&) {
+    remove_deleted_items(envi, deletedVarDecls);
+  } catch (ModelInconsistent&) { /* NOLINT(bugprone-empty-catch) */
   }
 }
 
 class SubstitutionVisitor : public EVisitor {
 protected:
   std::vector<VarDecl*> _removed;
+
+public:
   Expression* subst(Expression* e) {
     if (auto* vd = Expression::dynamicCast<VarDecl>(follow_id_to_decl(e))) {
       if (vd->type().isbool() && (vd->ti()->domain() != nullptr)) {
@@ -1172,8 +1180,6 @@ protected:
     }
     return e;
   }
-
-public:
   /// Visit array literal
   void vArrayLit(ArrayLit* al) {
     for (unsigned int i = 0; i < al->size(); i++) {
@@ -1236,7 +1242,10 @@ void substitute_fixed_vars(EnvI& env, Item* ii, std::vector<VarDecl*>& deletedVa
     }
   } else {
     auto* si = ii->cast<SolveI>();
-    top_down(sv, si->e());
+    if (si->e() != nullptr) {
+      si->e(sv.subst(si->e()));
+      top_down(sv, si->e());
+    }
     for (ExpressionSetIter it = si->ann().begin(); it != si->ann().end(); ++it) {
       top_down(sv, *it);
     }
@@ -1523,7 +1532,22 @@ bool simplify_constraint(EnvI& env, Item* ii, std::vector<VarDecl*>& deletedVarD
           } else {
             CollectDecls cd(env, env.varOccurrences, deletedVarDecls, ii);
             top_down(cd, c);
-            ii->remove();
+            if (auto* vdi = ii->dynamicCast<VarDeclI>()) {
+              auto* v = IntLit::a(b2i_val ? 1 : 0);
+              if (env.varOccurrences.occurrences(vdi->e()) == 0) {
+                if (is_output(vdi->e())) {
+                  VarDecl* vd_out = (*env.output)[env.outputFlatVarOccurrences.find(vdi->e())]
+                                        ->cast<VarDeclI>()
+                                        ->e();
+                  vd_out->e(v);
+                }
+                vdi->remove();
+              } else {
+                vdi->e()->e(v);
+              }
+            } else {
+              ii->remove();
+            }
           }
         } else {
           Id* ident = Expression::cast<Id>(c->arg(0));
@@ -2074,4 +2098,23 @@ void simplify_bool_constraint(EnvI& env, Item* ii, VarDecl* vd, bool& remove,
   }
 }
 
+void substitute_fixed_vars(Env& env) {
+  env.envi().checkCancel();
+  if (env.envi().failed()) {
+    return;
+  }
+  try {
+    EnvI& envi = env.envi();
+    Model& m = *envi.flat();
+    std::vector<VarDecl*> deletedVarDecls;
+    for (auto* item : m) {
+      if (!item->removed()) {
+        substitute_fixed_vars(envi, item, deletedVarDecls);
+      }
+    }
+    remove_deleted_items(envi, deletedVarDecls);
+  } catch (ModelInconsistent&) { /* NOLINT(bugprone-empty-catch) */
+  }
+}
+
 }  // namespace MiniZinc
