@@ -279,8 +279,6 @@ public:
     if (c->decl() != nullptr) {
       if (c->decl()->e() != nullptr || (c->argCount() == 1 && c->id() == "enum_of")) {
         if (c->decl()->isPolymorphic()) {
-          assert(c->argCount() != 1 || Expression::type(c->arg(0)).st() == Type::ST_PLAIN ||
-                 Expression::type(c->arg(0)).ot() == Type::OT_PRESENT);
           agenda.push(c);
         }
       }
@@ -359,7 +357,7 @@ public:
       curType.typeId(concrete_type.typeId());
       ti->type(curType);
       if (TIId* tiid = Expression::dynamicCast<TIId>(ti->domain())) {
-        ti_map.emplace(tiid->v(), curType);
+        ti_map.emplace(tiid->v(), ti->ranges().empty() ? curType : curType.elemType(env));
         if (curType.typeId() == 0) {
           // replace tiid with empty domain
           ti->domain(nullptr);
@@ -447,6 +445,9 @@ public:
       }
     } else if (TIId* tiid = Expression::dynamicCast<TIId>(ti->domain())) {
       Type ret_type = ti_map.find(tiid->v())->second;
+      if (ret_type.dim() != 0 && ti->type().dim() == 0) {
+        ret_type = ret_type.elemType(env);
+      }
       if (ret_type.structBT()) {
         ti->setStructDomain(env, ret_type, false, false);
       } else {
